/******************************************************************************
***  File: dm9kdiag.c version 1.00      				 		      ***
***  (C)Copyright 2000~2004 DAVICOM Semiconductor, Inc. All rights reserved.*** 
***														  	 ***	
***														  	 ***
***  Description:												 ***
***		This diagnostic test Register EEPROM Memory and TX/RX	  	  	 ***
***		of DM9000/DM9010/DM9000A 							  	 ***
***														  	 ***
***		Register r/w: read and write value from/to Register			 ***
***		EEPROM r/w  : read and write value from/to EEPROM		      	 ***
***		Memory	r/w : read and write value from/to SRAM	           	 ***
***		TX/RX test  : transmit and receive packet with lookback mode 	 ***	
***							  							 	 ***
***		Usage: dm9kdiag [-ewrmlch] [-p <IO port>] [-c <1 ~ 9999>]    	 ***
***															 ***
*******************************************************************************
*/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <getopt.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <ctype.h>


#if defined(__linux__) && __GNU_LIBRARY__ == 1
#include <asm/io.h>
#else			/* Newer libraries use <sys/io.h> instead. */
#include <sys/io.h>
#endif

typedef u_int32_t u32;
typedef u_int16_t u16;
typedef u_int8_t  u8;
typedef unsigned char uchar;

const char * const MEM_ERR_MSG =	
	"==== Write in memory pattern: %x ====\n"
	"==== read out data: %x		    ====\n\n";
const char * const MEM_PTR_MSG =
	"==== Current memory pointer: %x ====\n"
	"==== Correct memory pointer: %x ====\n\n";

const char * const REG_ERR_MSG =
	"==== Register[%d] test fail!! 	   ====\n"
	"==== Write in register pattern: %x   ====\n"
	"==== read out data: %x	  	        ====\n\n";

const char * const EEP_ERR_MSG =
	"==== EEPROM[%d] test fail!! 		 ====\n"
	"==== Write in memory pattern: %x   ====\n"
	"==== read out data: %x		   	 ====\n\n";

void ptrmsg(const char *fmt, u16 regxy, u16 memptr)
{
	printf(fmt, regxy, memptr);
} 
void errmsg(const char *fmt, int idx, u32 pat, u32 val)
{
	printf(fmt, idx, pat, val);
}

#define DM9K_NCR		0x00
#define DM9K_NSR		0x01
#define DM9K_RXCR		0x05
#define DM9K_REG05		0x30	/* SKIP_CRC/SKIP_LONG */
#define DM9K_REG08		0x37
#define DM9K_REG09		0x38
#define DM9K_REG0A		0x29	/* Flow Control Enable */ 
#define DM9K_GPCR		0x1e	/* General purpose control register */
#define DM9K_GPR		0x1f	/* General purpose register */
#define DM9K_REGFF		0x83	/* IMR */

#define DM9K_PHY		0x40	/* PHY address 0x01 */
#define DM9K_PKT_MAX		1536	/* Received packet max size */
#define DM9K_PKT_RDY		0x01	/* Packet ready to receive */
#define DM9K_MIN_IO		0x300
#define DM9K_MAX_IO		0x370

#define MAX_PACKET_SIZE 	1514
#define DMFE_MAX_MULTICAST 	14

#define DM9K_RX_INTR		0x01
#define DM9K_TX_INTR		0x02
#define DM9K_OVERFLOW_INTR	0x04

#define TRUE			1
#define FALSE			0

#define DM9000		0x9000
#define DM9010		0x9010

#define MAXSIZE		9999
#define PACKETLENS	1540

#define DIM(x)		(sizeof(x)/sizeof(x[0]))




#if (IO_MODE == 1)
#define	MEM_W(x, y)	outb(x, y)
#define 	MEM_R(x)		inb(x)
#define	MEMSIZE		DIM(x)
typedef	uchar	MEM_TYPE;
#elif (IO_MODE == 2)
#define	MEM_W(x, y)	outw(x, y)
#define	MEM_R(x)		inw(x)
typedef	u16	MEM_TYPE;
#elif (IO_MODE == 4)
#define	MEM_W(x, y)	outl(x, y)
#define	MEM_R(x)		inl(x)
#define	MEMSIZE		((DIM(x) + 3) / 4);
typedef	u32	MEM_TYPE;
#else
#error *** You should specify access mode
#endif

#pragma pack(push, 1)
typedef struct _RX_DESC
{
		u8	own;
		u8	status;
		u16 length;
} RX_DESC;

typedef union {
		u8	buf[4];
		RX_DESC	desc;
} rx_t;
#pragma pack(pop)

static int eepromrw();
extern void registerrw();
extern void memoryrw();
extern void lookbacktr();
extern void set_phy_mode();
extern void set_init();
static u8 ior(u8);
static void iow(u8, int);
static u16 eepread(int);
static void eepwrite(u16 , int);

static char *usage_msg = 
"Usage: dm9kdiag -ewrmlch [-p <IO port>] [-c <1 ~ 9999>]\n";

static const char * const long_usage_msg = 
"Usage: dm9kdiag -ewrmlch [-p <IO port>] [-c <1 ~ 9999>]\n"
"\n"
"Frequently used options are						   \n"
"-p	--ioaddr			Set new IO base address\n"
"-e 	--eread				Read value from EEPROM\n"
"-w	--ewrite			Write value into EEPROM\n"
"-r	--Regrw				Read/write value from/into Register\n"
"-m 	--Memrw				Read/write value from/into SRAM\n"
"-l	--Lookback			Send one packet out and receive by Lookback cable\n"
"-c 	--Continuous			Send multiple packet out and receive by Lookback cable\n"
"-h	--Help\n";

struct option longopts[] = {
/* { name has_arg *flag val } */
	{"base-address", 1, 0, 'p'},
	{"eepr", 0, 0, 'e'},
	{"eepw", 0, 0, 'w'},
	{"regrw", 0, 0, 'r'},
	{"memrw", 0, 0, 'm'},
	{"lookback", 0, 0, 'l'},
	{"continuous", 1, 0, 'c'},
	{"io-mode", 1, 0, 'i'},
	{"help", 0, 0, 'h'},
	{ 0, 0, 0, 0 }
};

char * chip_info[] = {
	"DM9000 ISA Ethernet Adapt",
	"DM9010 Ethernet Adapt",
	"DM9000A ISA Ethernet Adapt",
};

static int _geeprw		= FALSE;
static int _gwritein	= FALSE;
static int _gregrw		= FALSE;
static int _gmemrw		= FALSE;
static int _glookback	= FALSE;
static int _gnumbers	= 0;
static int _gdataaddr;
static int _gioaddr;
static u16 _gchipver, _gchiptype;


int main(int argc, char **argv)
{
	int c, longind;
	int port_base ;
	int new_ioaddr = -1;
	int errflag = 0;
	int helpflag = 0;
	int new_port = 0;
	u8 iomode;
		

	extern char *optarg;
	
	while((c = getopt_long(argc, argv, "ewrmlc:p:hi:", longopts, &longind)) != -1)
	{
		switch(c)
		{
			case 'e': _geeprw = TRUE;			break;
			case 'w': _geeprw = TRUE; _gwritein = TRUE;		break;
			case 'r': _gregrw = TRUE;			break;
			case 'm': _gmemrw = TRUE;			break;
			case 'l': _glookback = TRUE;			break;
			case 'c': 	
				  _gnumbers = strtol(optarg, NULL, 10);
				  printf("Input number is: %d\n", _gnumbers);

				  if (_gnumbers > 10000)
				  {
				  	fprintf(stderr, "To many number. Valid number (0 - 9999)\n");
				  	_gnumbers = 0;
				  }
				break;
			case 'p': 
					new_port = TRUE;
					new_ioaddr = strtol(optarg, NULL, 16);
					printf("new_ioaddr: %x\n", new_ioaddr);
					if (new_ioaddr < DM9K_MIN_IO || new_ioaddr > DM9K_MAX_IO)
					{
						fprintf(stderr, "Invalid new I/O addrex %#x. Valid range <0x300 - 0x370>\n", new_ioaddr);
						new_port = FALSE;
					}
					break;
			case 'h': helpflag = TRUE;		break;
			case '?':	
			default:
					errflag = TRUE;
		}
		if (helpflag)
		{
			fprintf(stderr, long_usage_msg);
			return 0;
		}

		if (errflag)
		{
			fprintf(stderr, usage_msg);
			return 3;
		}
	}

	if (new_port)
	{
		
		/* Set new I/O port */
		if (iopl(3))
		{
			fprintf(stderr, "I/O perm. error\n");
			return 2;
		}
		printf("Set I/O port success\n");
		_gioaddr = new_ioaddr;
	}
	
	if (!new_port)
	{
		port_base = DM9K_MIN_IO;

		if (iopl(3))
		{
			printf("I/O space access error!\n");
			return 2;
		}
		_gioaddr = port_base;
	}
	printf("I/O space access success:I/O address: %x\n\n", _gioaddr);
	_gdataaddr = _gioaddr + 4;
	iomode = (ior(0xfe) >> 6);
	if (iomode == 0)
		iomode = 8 *2;
	else if(iomode == 1)
		iomode = 8 * 4;
	else
		iomode = 8 * 1;
	printf("====This adaptor running in %d-bits mode,====\n", iomode);
	printf("====the command must be IO_MODE=%d in Makefile\n\n", (iomode/8));

	_gchipver = ior(0x2c);
	printf("chip version is %02X\n", _gchipver);
	if (_gchipver == 0x00)
	{
		_gchiptype = 0x9000;
		printf("<<%s>>\n", *(chip_info));
	}	
	else if (_gchipver == 0x10)
	{
		_gchiptype = 0x9010;
		printf("<<%s>>\n", *(chip_info+1));
	}
	else if (_gchipver == 0x18)
	{
		_gchiptype = 0x900a;
		printf("<<%s>>\n", *(chip_info+2));
	}
	else
		printf("<<Unknow chip type>>\n");

	if (_geeprw)
	do{
		int tmp;
		tmp = eepromrw();
		_gnumbers--;
	}while(_gnumbers > 0);
		
	if (_gregrw)
	do{
		registerrw();
		_gnumbers--;
	}while(_gnumbers > 0);
			
	if(_gmemrw)
	do{
		memoryrw();
		_gnumbers--;
	}while(_gnumbers > 0);
			
	if(_glookback)
	do{	
		lookbacktr();
		_gnumbers--;
	}while(_gnumbers > 0);
	
	return 0;
}

static int eepromrw()
{
	const int EPSIZE = 12;
	u16 recvdata[EPSIZE], tmpdata[EPSIZE];
	u16 testpattern[6] = {0xff00, 0x00ff, 0x5aa5, 0xa55a, 0x55aa, 0xaa55};
	u16 getdata;
	
	int pindex, offset, err = 0;
	

	memset(&recvdata, 0, sizeof(recvdata));
	
	if(_gwritein)
	{
		int i, j;
		/* Read data into tmparray from EEPROM  */
		for (i = 0; i <= 12; i++)
			tmpdata[i] = eepread(i);
		
		for (offset = 0; offset <= DIM(recvdata); offset++)
		{
			printf("==== EEPROM[%d] test ====\n", offset);
			for (pindex = 0; pindex < DIM(testpattern); pindex++)
			{
				eepwrite(testpattern[pindex], offset); // write pattern into EEPROM
				getdata = eepread(offset);	// read data from EEPROM
				printf("==== Write in value: %x \n", testpattern[pindex]);
				if (getdata == testpattern[pindex])
					continue;
				/* Test fail */
				err++;
				errmsg(EEP_ERR_MSG, offset, 
					testpattern[pindex], getdata);
			} /* end of for (pindex = 0; 
				pindex < DIM(testpattern); pindex++) */
		} //end of for (offset = 0; offset <= 12; offset++)
		if (!err)
			printf("==== EEPROM test pass!! ====\n");

		/* write original value back to EEPROM */	
		for (j = 0; j <= DIM(tmpdata); j++)
			eepwrite(tmpdata[j], j);

		return 1;
	} //end of if(_gwritein)
	
	
	/* Read data from EEPROM */
	printf("*******************************\n");
	printf("**** Read data from EEPROM ****\n");
	for (offset = 0; offset < DIM(recvdata); offset++)
	{
		getdata = eepread(offset);
		getdata = (ior(0xd) + (ior(0xe) << 8));
		printf("**** EEPROM[%x]: %x        ****\n", offset, getdata);
	}
	printf("*******************************\n");
	return 0;
}

void registerrw(void)
{
	//u16 regvalue;
	
	uchar dm9reg[] = {0x10, 0x11, 0x12, 0x13, 0x14, 0x15};
	uchar testpattern[] = { 0xff, 0x00, 0x5a, 0xa5, 0x55, 0xaa};
	uchar retvalue;

	int regloop, pindex, err = 0;
		
	printf("===========================================================\n");
	printf("====== Register(0x10 ~ 0x15) read and wriet for test ======\n");
	printf("===========================================================\n");

	for (regloop = 0; regloop < DIM(dm9reg); regloop++)
	{	
		printf("==== Test Reg[%2x] ====\n", dm9reg[regloop]);
		
		//int j;
		for (pindex = 0; pindex < DIM(testpattern); pindex++)
		{
			
			iow(dm9reg[regloop], testpattern[pindex]);
			retvalue = ior(dm9reg[regloop]);
			printf("==== Write in value: %x\n", retvalue);
			
			if (retvalue == testpattern[pindex])	
				continue;
			
			/* test fail */
			err++;
			errmsg(REG_ERR_MSG, dm9reg[regloop], 
				(u32)testpattern[pindex], (u32)retvalue);
 
		} // end of for(j = 0; j < DIM(testpattern); j++)
	} // end of for(regloop = 0; regloop < DIM(dm9reg); regloop++)
	if (!err)
		printf("==== Register test pass!! ====\n");
	printf("\n");

}

void memoryrw(void)
{
	int offset, pindex;
	int err = 0;

	uchar testpattern[] = { 0xff, 0x00, 0x5a, 0xa5, 0x55, 0xaa, 0xaa, 0x55};
	//uchar retvalue;

	MEM_TYPE *ptrpattern = (MEM_TYPE*)testpattern;
	volatile MEM_TYPE	retvalue;
	u16 memptr = 0, regab, reg45;

	u16 tmpsize, memsize = 4096;
	//u16 tmpsize;

	
	printf("******************************************\n");
	printf("****** Memory read & write for test ******\n");
	printf("******************************************\n");
	
	/* Initial Register 0xf4, 0xf5, 0xfa, 0xfb, 0xff */
	iow(0xf4, 0);
	iow(0xf5, 0);
	iow(0xfa, 0);
	iow(0xfb, 0);
	iow(0xff, 0);


	err = 0;
	printf("==== %d-bits mode test ====\n",(IO_MODE * 8));
	tmpsize = (DIM(testpattern) + IO_MODE - 1) / IO_MODE;
	printf("<,, %d>\n", tmpsize);
	for (pindex = 0; pindex < tmpsize; pindex++)
	{	
		for (offset = 0; offset < memsize; offset++)
		{
			/* Move data to SRAM & dymmy read */
			memptr = (memptr + IO_MODE) % 0x4000;
			//printf("<w,%x,%x>", memptr, *(ptrpattern+pindex));
			//fflush(stdout);
			//usleep(5000);
			memptr++;
			outb(0xf8, _gioaddr);
			MEM_W(*(ptrpattern + pindex), _gdataaddr);
			regab = (ior(0xfb) << 8) + (ior(0xfa));

			/* Read data from SRAM */
			outb(0xf0, _gioaddr);
			MEM_R(_gdataaddr);
			outb(0xf2, _gioaddr);

			reg45 = (ior(0xf5) << 8) + (ior(0xf4));
			retvalue = MEM_R(_gdataaddr);
			
			if (retvalue == *(ptrpattern + pindex))
				continue;
				
			err++;
			
			if (memptr != regab)
			{
				printf("\n");
				printf("==== Memory pointer error after write value into SRAM ====\n");
				ptrmsg(MEM_PTR_MSG, regab, memptr);
			}
			if (memptr != reg45)
			{
				printf("==== Memory pointer error after read value from SRAM ====\n");
				ptrmsg(MEM_PTR_MSG, reg45, memptr);
			}
			
			if (retvalue != *(ptrpattern + pindex))
				errmsg(MEM_ERR_MSG, 
					(u32)testpattern[pindex], (u32)retvalue, pindex);
				
			if ( ++err > 3) return ;
		}//end of for(pindex=0; pindex <= DIM(testpattern; pindex++)
	}// end of for (offset = 0; offset <= tmpsize; offset++)
	if (!err)
		printf("==== Memory test in %d-bits mode pass!! ====\n", (IO_MODE * 8));
}

void lookbacktr(void)
{
	int err, i;
	//int pktsize = MAX_PACKET_SIZE;
	int pktsize = 258;
	int len = (pktsize + IO_MODE - 1) / IO_MODE;
	
	u8 reg02;
	u8 intrstats;
	
	u8 txdata[MAX_PACKET_SIZE], rxdata[MAX_PACKET_SIZE];

	MEM_TYPE *txptr = (MEM_TYPE*)txdata;
	MEM_TYPE *rxptr = (MEM_TYPE*)rxdata;

	memset(&txdata, 0, MAX_PACKET_SIZE);
	memset(&rxdata, 0, MAX_PACKET_SIZE);
	
	set_phy_mode();
	set_init();
	
	/* Disable all interrupt */
	iow(0xff, 0x80);

	outb(0xf8, _gioaddr);

	memset(&txdata, 0xff, 6);
	for (i = 6; i <= len; i++)
		*(txptr+i) = (MEM_TYPE)i;

	printf("\n");
	printf("packet length: %d\n", len);

	printf("<< <TX> %d-bits mode>>\n", (IO_MODE*8));
	for (i = 0; i <= len; i++)
		MEM_W(*(txptr+i), _gdataaddr);

	printf("\n");
		
	iow(0xfc, pktsize & 0xff);
	iow(0xfd, (pktsize >> 8) & 0xff);
			
	/* issue TX polling command */
	iow(0x2, 0x1);
	do{
		reg02 = ior(0x02);		// Cleared after TX complete */
		usleep(30);
	}while((reg02 & 1));


	intrstats = ior(0xfe);
	iow(0xfe, intrstats);
	if ((intrstats & 0x01))
		printf("==== RX interrupt detected ====\n\n");
	if (intrstats & 0x02)
		printf("==== TX interrupt detected ====\n\n");
		
	rx_t x;

	/* dummy read */
	outb(0xf0, _gioaddr);
	MEM_R(_gdataaddr);
			
	MEM_TYPE * ptr = (MEM_TYPE*)&x;

	outb(0xf2, _gioaddr);
	for(i=4/IO_MODE;
		i--;
		*(ptr++) = MEM_R(_gdataaddr)
		);

	printf("<own:%02x>, <sta:%02x>, <len:%04x>\n",
		x.desc.own, x.desc.status, x.desc.length);

	/* minus 4 byte CRC */
	x.desc.length -= 4;

	if (x.desc.own != DM9K_PKT_RDY) 
	{
		printf("error own byte %02X\n", x.desc.own);
		return;
	}

	if(x.desc.length != pktsize)
	{
		printf("error in receive length = %d\n", x.desc.length);
		return;
	}

	if (x.desc.length < 0x40 || x.desc.length > DM9K_PKT_MAX)
	{
		printf("*** Bad packet length = %d\n", x.desc.length);
		return;
	}

	if (x.desc.status & 0xbf00)
	{

		if (x.desc.status & 0x100)
			printf("**** RX FIFO error ****\n");
		if (x.desc.status & 0x200)
			printf("**** RX CRC error ****\n");
		if (x.desc.status & 0x8000)
			printf("**** RX length error ****\n");

		return;
	}

	/* start to receive buffer */
	printf("<< <RX> %d-bits mode >>\n\n", (IO_MODE*8));

	/* dummy read */
	outb(0xf0, _gioaddr);
	MEM_R(_gdataaddr);

	outb(0xf2, _gioaddr);
	for(i = (x.desc.length+IO_MODE-1)/IO_MODE;
		i--;
		*(rxptr++) = MEM_R(_gdataaddr)); 

	err = 0;
	txptr = (MEM_TYPE*) &txdata;
	rxptr = (MEM_TYPE*) &rxdata;
#if 0	
	for(i = (x.desc.length+IO_MODE-1)/IO_MODE;
		i--;
		rxptr++, txptr++
		)
	{
		if(*rxptr == *txptr) continue;
		err++;
		printf("<%x>", (int)rxptr-(int)&rxdata);
	}
#endif
	for (i = 0; i < len; i++, rxptr++, txptr++)
	{
		if (*rxptr == *txptr) continue;
		err++;
		printf("<offset: %x>", i*IO_MODE);
	}
	if(err)
		printf("==== Receive packet error!! ====\n");
	else
		printf("==== Lookback test pass!! ====\n");
}

void set_phy_mode(void)
{
	u16 phy_reg0 = 0x1000;		/* Auto-negotiation & non-duplux mode*/
	u16 phy_reg4 = 0x01e1;		/* Default flow control disable */


	/* PHY write */
	/* write phy reg0 */

	/* Fill the phyxcer register into REG_0C */
	iow(0xc, DM9K_PHY | 0);

	/* Fill the written data into REG_0D & REG_0E */
	iow(0xd, (phy_reg0 & 0xff));
	iow(0xe, ( (phy_reg0 >> 8) & 0xff));

	iow(0xb, 0xa);		/* Issue phyxcer write command */
	usleep(500);			/* Wait write complete */
	iow(0xb, 0x0);		/* Clear phyxcer write command */
	
	/* write phy reg4 */
	iow(0xc, DM9K_PHY |4);
	iow(0xd, (phy_reg4 & 0xff));
	iow(0xe, ((phy_reg4 >> 8) & 0xff));
	iow(0xb, 0xa);
	usleep(500);
	iow(0xb, 0x0);
	
	
}

void set_init(void)
{

	/* set the internal PHY power-on, GPIOs normal, and wait 2ms */
	iow(0x1F, 0);	/* GPR (reg_1Fh)bit GPIO0=0 pre-activate PHY */
	usleep(20);		/* wait 2ms for PHY power-on ready */
	
	iow(0x00, 3);
	iow(0x02, 0);		/* TX Polling clear */
	iow(0x2f, 0);		/* Special Mode */
	iow(0x01, 0x2c);	/* clear TX status */
	iow(0xfe, 0x0f); 	/* Clear interrupt status */
	iow(0x08, 0x37);
	iow(0x09, 0x38);	/* Flow control: High/Low water */
	iow(0x0a, 0x29);	/* flow control */

	/* Set MAC address */
	iow(0x10, 0x00);
	iow(0x11, 0x60);
	iow(0x12, 0x6e);
	iow(0x13, 0x11);
	iow(0x14, 0x00);
	iow(0x15, 0x07);

	/* Set Multicast address */
	iow(0x16, 0xff);
	iow(0x17, 0xff);
	iow(0x18, 0xff);
	iow(0x19, 0xff);
	iow(0x1a, 0xff);
	iow(0x1b, 0xff);
	iow(0x1c, 0xff);
	iow(0x1d, 0xff);
	
	/* Activate DM9000 */
	iow(DM9K_RXCR, 0x01);
	iow(0xff, 0x83);	/* Enable TX/RX interrupt mask */
}

static u8 ior(u8 reg)
{
	outb(reg, _gioaddr);
	return inb(_gioaddr + 4);
}

static void iow(u8 reg, int value)
{
	outb(reg, _gioaddr);
	usleep(100);
	outb(value, _gioaddr + 4);
}

static u16 eepread(int offset)
{
	iow(0xc, offset);
	iow(0xb, 0x4);
	usleep(200);
	iow(0xb, 0);
	return (ior(0xd) + (ior(0xe) << 8));
}

static void eepwrite(u16 patternval, int offset)
{
	iow(0x0c, offset);
	iow(0x0d, patternval & 0xff);
	iow(0x0e, (patternval >> 8) & 0xff);
	iow(0x0b, 0x12);
	usleep(200);
	iow(0x0b, 0);
}

